# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from google.protobuf import message
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import os
import sys
from ccdexplorer_fundamentals.enums import NET
from typing import TYPE_CHECKING


sys.path.append(os.path.dirname("ccdefundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *

sys.path.append(os.path.dirname("ccdefundamentals"))
if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def convertKeysEntry(self, message) -> dict[str, CCD_AccountVerifyKey]:
        entry_dict = {}
        for index, entry in message.items():
            converted_value = self.convertType(entry)
            entry_dict[str(index)] = CCD_AccountVerifyKey(
                **{"ed25519_key": converted_value}
            )

        return entry_dict

    def convertArData(self, message) -> dict[str, CCD_ChainArData]:
        entry_dict = {}
        for index, entry in message.items():
            converted_value = self.convertType(entry)
            entry_dict[str(index)] = CCD_ChainArData(
                **{"enc_id_cred_pub_share": converted_value}
            )

        return entry_dict

    def convertAttributesEntry(self, message) -> dict[str, CCD_Commitment]:
        attributes = {}
        for element, value in message.items():
            converted_value = self.convertType(value)
            attributes[CredentialElement(element).name] = converted_value

        return attributes

    def convertPolicy_AttributesEntry(self, message) -> CCD_Policy_Attributes:
        attributes = {}
        for element, value in message.items():
            if CredentialElement(element).name == "idDocType":
                value = CredentialDocType(bytes.decode(value)).name.replace("_", " ")
                attributes[CredentialElement(element).name] = value

            else:
                attributes[CredentialElement(element).name] = bytes.decode(value)

        return attributes

    def convertCredentialPublicKeys(self, message) -> CCD_CredentialPublicKeys:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif key == "keys":
                result[key] = self.convertKeysEntry(value)

        return CCD_CredentialPublicKeys(**result)

    def convertYearMonth(self, message) -> CCD_YearMonth:
        result = self.convertTypeWithSingleValues(message)

        return CCD_YearMonth(**result)

    def convertPolicy(self, message) -> CCD_Policy:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if key == "attributes":
                result[key] = self.convertPolicy_AttributesEntry(value)

            elif type(value) == YearMonth:
                result[key] = self.convertYearMonth(value)

        return CCD_Policy(**result)

    def convertListOfCommitments(self, message) -> CCD_Commitment:
        entries = []

        for list_entry in message:
            entries.append(self.convertType(list_entry))

        return entries

    def convertCredentialCommitments(self, message) -> CCD_CredentialCommitments:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif key == "attributes":
                result[key] = self.convertAttributesEntry(value)

            elif key == "id_cred_sec_sharing_coeff":
                result[key] = self.convertListOfCommitments(value)

        return CCD_CredentialCommitments(**result)

    def convertInitialCredentialValues(self, message) -> CCD_InitialCredentialValues:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) == CredentialPublicKeys:
                result["credential_public_keys"] = self.convertCredentialPublicKeys(
                    value
                )

            elif type(value) == Policy:
                result[key] = self.convertPolicy(value)

        return CCD_InitialCredentialValues(**result)

    def convertNormalCredentialValues(self, message) -> CCD_NormalCredentialValues:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) == CredentialPublicKeys:
                result["credential_public_keys"] = self.convertCredentialPublicKeys(
                    value
                )

            elif key == "ar_data":
                result[key] = self.convertArData(value)

            elif type(value) == Policy:
                result[key] = self.convertPolicy(value)

            elif type(value) == CredentialCommitments:
                result[key] = self.convertCredentialCommitments(value)

        return CCD_NormalCredentialValues(**result)

    def convertAccountCredential(self, message) -> CCD_AccountCredential:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) == InitialCredentialValues:
                    result[key] = self.convertInitialCredentialValues(value)

                elif type(value) == NormalCredentialValues:
                    result[key] = self.convertNormalCredentialValues(value)

        return CCD_AccountCredential(**result)

    def convertCredentials(self, message) -> dict[str, CCD_AccountCredential]:
        cred_dict = {}
        for index, credential in message.items():
            cred_dict[str(index)] = self.convertAccountCredential(credential)

        return cred_dict

    def convertEncryptedBalance(self, message) -> CCD_EncryptedBalance:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif key == "incoming_amounts":
                result[key] = self.convertList(value)

        return CCD_EncryptedBalance(**result)

    def convertBakerInfo(self, message):
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return result

    def convertAccountStakingInfo(self, message: message):
        result = {}
        which_one = message.WhichOneof("staking_info")
        if not which_one:
            return CCD_AccountStakingInfo(**{"baker": None, "delegator": None})
        else:
            if which_one == "baker":
                for descriptor in getattr(message, which_one).DESCRIPTOR.fields:
                    key, value = self.get_key_value_from_descriptor(
                        descriptor, getattr(message, which_one)
                    )
                    if type(value) in self.simple_types:
                        result[key] = self.convertType(value)

                    elif type(value) == BakerPoolInfo:
                        result[key] = self.convertBakerPoolInfo(value)

                    elif type(value) == BakerInfo:
                        result[key] = self.convertBakerInfo(value)

                    elif type(value) == StakePendingChange:
                        result[key] = self.convertPendingChange(value)

            elif which_one == "delegator":
                for descriptor in getattr(message, which_one).DESCRIPTOR.fields:
                    key, value = self.get_key_value_from_descriptor(
                        descriptor, getattr(message, which_one)
                    )
                    if type(value) in [
                        BakerId,
                        AccountAddress,
                        Amount,
                        str,
                        int,
                        bool,
                        float,
                    ]:
                        result[key] = self.convertType(value)

                    elif type(value) == DelegationTarget:
                        result[key] = self.convertDelegationTarget(value)

                    elif type(value) == StakePendingChange:
                        result[key] = self.convertPendingChange(value)

            return CCD_AccountStakingInfo(**{which_one: result})

    def get_account_info(
        self: GRPCClient,
        block_hash: str,
        hex_address: str = None,
        account_index: int = None,
        net: Enum = NET.MAINNET,
    ) -> CCD_AccountInfo:
        blockHashInput = self.generate_block_hash_input_from(block_hash)
        if account_index is not None:
            accountIdentifierInput = (
                self.generate_account_identifier_input_from_account_index(account_index)
            )
        elif hex_address:
            accountIdentifierInput = self.generate_account_identifier_input_from(
                hex_address
            )

        account_info = AccountInfoRequest(
            block_hash=blockHashInput, account_identifier=accountIdentifierInput
        )

        grpc_return_value: AccountInfo = self.stub_on_net(
            net, "GetAccountInfo", account_info
        )

        result = {}
        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) == ReleaseSchedule:
                result[key] = self.convertRelease(value)

            elif type(value) == EncryptedBalance:
                result[key] = self.convertEncryptedBalance(value)

            # TODO
            # type EncryptedBalance
            # type EncryptionKey

            elif key == "creds":
                result["credentials"] = self.convertCredentials(value)

            elif type(value) == AccountStakingInfo:
                result[key] = self.convertAccountStakingInfo(value)

        return CCD_AccountInfo(**result)



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
from typing import Iterator
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import CCD_DelegatorRewardPeriodInfo
from typing import TYPE_CHECKING
import sys

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def get_delegators_for_passive_delegation_in_reward_period(
        self: GRPCClient,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> list[CCD_DelegatorRewardPeriodInfo]:
        result = []
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        self.check_connection(net, sys._getframe().f_code.co_name)
        if net == NET.MAINNET:
            grpc_return_value: Iterator[DelegatorInfo] = (
                self.stub_mainnet.GetPassiveDelegatorsRewardPeriod(
                    request=blockHashInput
                )
            )
        else:
            grpc_return_value: Iterator[DelegatorInfo] = (
                self.stub_testnet.GetPassiveDelegatorsRewardPeriod(
                    request=blockHashInput
                )
            )

        grpc_return_value: Iterator[DelegatorInfo] = self.stub_on_net(
            net, "GetPassiveDelegatorsRewardPeriod", blockHashInput
        )

        for delegator in list(grpc_return_value):
            result.append(
                CCD_DelegatorRewardPeriodInfo(
                    **{
                        "account": self.convertAccountAddress(delegator.account),
                        "stake": self.convertAmount(delegator.stake),
                    }
                )
            )

        return result



# ruff: noqa :E501, F405
import datetime as dt
from enum import Enum
from typing import Dict, Optional, Union

import motor.motor_asyncio
from motor.motor_asyncio import AsyncIOMotorCollection
from pydantic import BaseModel, ConfigDict, Field
from pymongo import MongoClient
from pymongo.collection import Collection
from rich.console import Console

from ccdexplorer_fundamentals.env import MONGO_URI
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *  # noqa: F403
from ccdexplorer_fundamentals.mongodb_queries._apy_calculations import (
    Mixin as _apy_calculations,
)
from ccdexplorer_fundamentals.mongodb_queries._baker_distributions import (
    Mixin as _distributions,
)
from ccdexplorer_fundamentals.mongodb_queries._search_transfers import (
    Mixin as _search_transfers,
)
from ccdexplorer_fundamentals.mongodb_queries._store_block import Mixin as _store_block
from ccdexplorer_fundamentals.mongodb_queries._subscriptions import (
    Mixin as _subscriptions,
)
from ccdexplorer_fundamentals.tooter import Tooter, TooterChannel, TooterType

console = Console()


class MongoLabeledAccount(BaseModel):
    id: str = Field(..., alias="_id")
    account_index: Optional[CCD_AccountIndex] = None
    label: str
    label_group: str


class AccountStatementEntry(BaseModel):
    block_height: int
    slot_time: dt.datetime
    entry_type: str
    amount: microCCD
    balance: microCCD


class AccountStatementTransferType(BaseModel):
    amount: microCCD
    counterparty: str


class AccountStatementEntryType(BaseModel):
    amount_decrypted: Optional[microCCD] = None
    amount_encrypted: Optional[microCCD] = None
    baker_reward: Optional[microCCD] = None
    finalization_reward: Optional[microCCD] = None
    foundation_reward: Optional[microCCD] = None
    transaction_fee: Optional[microCCD] = None
    transaction_fee_reward: Optional[microCCD] = None
    transfer_in: Optional[list[AccountStatementTransferType]] = None
    transfer_out: Optional[list[AccountStatementTransferType]] = None


class MongoImpactedAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(..., alias="_id")
    tx_hash: Optional[str] = None  # Rewards do not have a tx
    impacted_address: str
    impacted_address_canonical: str
    effect_type: str
    balance_movement: Optional[AccountStatementEntryType] = None
    block_height: int
    included_in_flow: Optional[bool] = None
    date: Optional[str] = None


class MongoTokensImpactedAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(..., alias="_id")
    tx_hash: str
    impacted_address: str
    impacted_address_canonical: str
    event_type: str
    token_address: str
    block_height: int
    date: str


class MongoTypeBlockPerDay(BaseModel):
    """
    Block Per Day. This type is stored in the collection `blocks_per_day`.

    :Parameters:
    - `_id`: the date of the day that ended
    - `date`: the date of the day that ended
    - `height_for_first_block`: height of the first block in the day
    - `height_for_last_block`: height of the last block in the day
    - `slot_time_for_first_block`: time of the first block in the day
    - `slot_time_for_last_block`: time of the last block in the day
    - `hash_for_first_block`: hash of the first block in the day
    - `hash_for_last_block`: hash of the last block in the day

    """

    id: str = Field(..., alias="_id")
    date: str
    height_for_first_block: int
    height_for_last_block: int
    slot_time_for_first_block: dt.datetime
    slot_time_for_last_block: dt.datetime
    hash_for_first_block: str
    hash_for_last_block: str


class MongoTypeInvolvedAccount(BaseModel):
    """
    Involved Account. This type is stored in the collections `involved_accounts_all` and
    `involved_accounts_transfer`.

    :Parameters:
    - `_id`: the hash of the transaction
    - `sender`: the sender account address
    - `receiver`: the receiver account address, might be null
    - `sender_canonical`: the canonical sender account address
    - `receiver_canonical`: the  canonical receiver account address, might be null
    - `amount`: amount of the transaction, might be null
    - `type`: dict with transaction `type` and `contents`
    - `block_height`: height of the block in which the transaction is executed
    """

    id: str = Field(..., alias="_id")
    sender: str
    receiver: Optional[str] = None
    sender_canonical: str
    receiver_canonical: Optional[str] = None
    amount: Optional[int] = None
    type: dict[str, str]
    block_height: int
    memo: Optional[str] = None


class MongoTypeInvolvedContract(BaseModel):
    """
    Involved Contract. This type is stored in the collection `involved_contracts`.

    :Parameters:
    - `_id`: the hash of the transaction - the `str` representation of the contract.
    - `index`: contract index
    - `subindex`: contract subindex
    - `contract`: the `str` representation of the contract
    - `type`: dict with transaction `type` and `contents`
    - `block_height`: height of the block in which the transaction is executed
    - `source_module`: hash of the source module from which this contract is instanciated.
    """

    id: str = Field(..., alias="_id")
    index: int
    subindex: int
    contract: str
    type: dict[str, str]
    block_height: int
    source_module: str


class MongoTypeModule(BaseModel):
    """
    Module. This type is stored in the collection `modules`.

    :Parameters:
    - `_id`: the hex string
    - `module_name`: the name from the module
    - `methods`: list of method names
    - `contracts`: list of contract instances from this module
    """

    id: str = Field(..., alias="_id")
    module_name: str
    methods: Optional[list[str]] = None
    contracts: Optional[list[str]] = None
    init_date: Optional[dt.datetime] = None


class MongoTypeInstance(BaseModel):
    """
    Instance. This type is stored in the collection `instances`.

    :Parameters:
    - `_id`: the hex string
    - `module_name`: the name from the module
    - `methods`: list of method names
    - `contracts`: list of contract instances from this module
    """

    id: str = Field(..., alias="_id")
    v0: Optional[CCD_InstanceInfo_V0] = None  # noqa: F405
    v1: Optional[CCD_InstanceInfo_V1] = None  # noqa: F405


class MongoTypeReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`, property `reward`.

    """

    pool_owner: Optional[Union[int, str]] = None
    account_id: Optional[str] = None
    transaction_fees: int
    baker_reward: int
    finalization_reward: int


class MongoTypePoolReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`.

    :Parameters:
    - `_id`: the hex string

    """

    id: str = Field(..., alias="_id")
    pool_owner: Union[int, str]
    pool_status: dict
    reward: MongoTypeReward
    date: str
    slot_time: dt.datetime


class MongoTypeAccountReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`.

    :Parameters:
    - `_id`: the hex string
    - `module_name`: the name from the module
    - `methods`: list of method names
    - `contracts`: list of contract instances from this module
    """

    id: str = Field(..., alias="_id")
    account_id: str
    staked_amount: int
    account_is_baker: Optional[bool] = None
    baker_id: Optional[int] = None
    reward: MongoTypeReward
    date: str
    slot_time: dt.datetime


class Delegator(BaseModel):
    account: str
    stake: int


class MongoTypePayday(BaseModel):
    """
    Payday. This type is stored in collection `paydays`.

    :Parameters:
    - `_id`: hash of the block that contains payday information for
    this payday.
    - `date`: the payday date
    - `height_for_first_block`: height of the first block in the payday
    - `height_for_last_block`: height of the last block in the payday
    - `hash_for_first_block`: hash of the first block in the payday
    - `hash_for_last_block`: hash of the last block in the payday
    - `payday_duration_in_seconds`: duration of the payday in seconds (used for
    APY calculation)
    - `payday_block_slot_time`: time of payday reward block
    - `bakers_with_delegation_information`: bakers with delegators for reward period, retrieved
    from `get_delegators_for_pool_in_reward_period`, using the hash of the last block
    - `baker_account_ids`: mapping from baker_id to account_address
    - `pool_status_for_bakers`: dictionary, keyed on pool_status, value
    is a list of bakers, retrieved using the hash of the first block
    """

    id: str = Field(..., alias="_id")
    date: str
    height_for_first_block: int
    height_for_last_block: int
    hash_for_first_block: str
    hash_for_last_block: str
    payday_duration_in_seconds: float
    payday_block_slot_time: dt.datetime
    bakers_with_delegation_information: dict[str, list[Delegator]]
    baker_account_ids: dict[int, str]
    pool_status_for_bakers: Optional[dict[str, list[int]]] = None


class MongoTypePaydayAPYIntermediate(BaseModel):
    """
    Payday APY Intermediate. This type is stored in collection `paydays_apy_intermediate`.

    :Parameters:
    - `_id`: baker_is or account address

    """

    id: str = Field(..., alias="_id")
    daily_apy_dict: dict
    d30_apy_dict: Optional[dict] = None
    d90_apy_dict: Optional[dict] = None
    d180_apy_dict: Optional[dict] = None


class MongoTypePaydaysPerformance(BaseModel):
    """
    Payday Performance. This is a collection that stores daily performance characteristics
    for bakers.


    :Parameters:
    - `_id`: unique id in the form of `date`-`baker_id`
    - `expectation`: the daily expected number of blocks for this baker in this payday.
    Calculated as the lottery power * 8640 (the expected number of blocks in a day)
    - `payday_block_slot_time`: Slot time of the payday block
    - `baker_id`: the baker_id
    - `pool_status`:

    """

    id: str = Field(..., alias="_id")
    pool_status: CCD_PoolInfo  # noqa: F405
    expectation: float
    date: str
    payday_block_slot_time: dt.datetime
    baker_id: str


class Collections(Enum):
    blocks = "blocks"
    transactions = "transactions"
    special_events = "special_events"
    instances = "instances"
    modules = "modules"
    messages = "messages"
    paydays = "paydays"
    paydays_performance = "paydays_performance"
    paydays_rewards = "paydays_rewards"
    paydays_apy_intermediate = "paydays_apy_intermediate"
    paydays_current_payday = "paydays_current_payday"
    paydays_helpers = "paydays_helpers"
    involved_accounts_all = "involved_accounts_all"
    involved_accounts_all_top_list = "involved_accounts_all_top_list"
    involved_accounts_transfer = "involved_accounts_transfer"
    involved_contracts = "involved_contracts"
    nightly_accounts = "nightly_accounts"
    blocks_at_end_of_day = "blocks_at_end_of_day"
    blocks_per_day = "blocks_per_day"
    helpers = "helpers"
    memo_transaction_hashes = "memo_transaction_hashes"
    cns_domains = "cns_domains"
    bot_messages = "bot_messages"
    dashboard_nodes = "dashboard_nodes"
    tokens_accounts = "tokens_accounts"
    tokens_links_v2 = "tokens_links_v2"
    tokens_token_addresses_v2 = "tokens_token_addresses_v2"
    tokens_tags = "tokens_tags"
    tokens_logged_events = "tokens_logged_events"
    tokens_token_addresses = "tokens_token_addresses"
    memos_to_hashes = "memos_to_hashes"
    credentials_issuers = "credentials_issuers"
    impacted_addresses = "impacted_addresses"
    impacted_addresses_pre_payday = "impacted_addresses_pre_payday"
    impacted_addresses_all_top_list = "impacted_addresses_all_top_list"
    # statistics and pre-renders
    pre_tokens_overview = "pre_tokens_overview"
    pre_addresses_by_contract_count = "pre_addresses_by_contract_count"
    pre_tokens_by_address = "pre_tokens_by_address"
    statistics = "statistics"
    pre_render = "pre_render"

    # addresses and contracts per net per usecase
    usecases = "usecases"
    tokens_impacted_addresses = "tokens_impacted_addresses"


class CollectionsUtilities(Enum):
    labeled_accounts = "labeled_accounts"
    labeled_accounts_metadata = "labeled_accounts_metadata"
    users_prod = "users_prod"
    users_dev = "users_dev"
    exchange_rates = "exchange_rates"
    exchange_rates_historical = "exchange_rates_historical"
    users_v2_prod = "users_v2_prod"
    users_v2_dev = "users_v2_dev"
    message_log = "message_log"
    preferences_explanations = "preferences_explanations"
    release_notes = "release_notes"
    token_api_translations = "token_api_translations"
    # use case management
    usecases = "usecases"
    helpers = "helpers"


class MongoDB(
    _search_transfers,
    _subscriptions,
    _distributions,
    _store_block,
    _apy_calculations,
):
    def __init__(self, tooter: Tooter):
        self.tooter: Tooter = tooter
        try:
            con = MongoClient(MONGO_URI)
            self.connection: MongoClient = con

            self.mainnet_db = con["concordium_mainnet"]
            self.mainnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.mainnet[collection] = self.mainnet_db[collection.value]

            self.testnet_db = con["concordium_testnet"]
            self.testnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.testnet[collection] = self.testnet_db[collection.value]

            self.devnet_db = con["concordium_devnet"]
            self.devnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.devnet[collection] = self.devnet_db[collection.value]

            self.utilities_db = con["concordium_utilities"]
            self.utilities: Dict[CollectionsUtilities, Collection] = {}
            for collection in CollectionsUtilities:
                self.utilities[collection] = self.utilities_db[collection.value]

            console.log(con.server_info()["version"])
        except Exception as e:
            print(e)
            tooter.send(
                channel=TooterChannel.NOTIFIER,
                message=f"BOT ERROR! Cannot connect to MongoDB, with error: {e}",
                notifier_type=TooterType.MONGODB_ERROR,
            )


class MongoMotor(
    _search_transfers,
    _subscriptions,
    _distributions,
    _store_block,
    _apy_calculations,
):
    def __init__(self, tooter: Tooter):
        self.tooter: Tooter = tooter
        try:
            con = motor.motor_asyncio.AsyncIOMotorClient(MONGO_URI)
            self.connection = con

            self.mainnet_db = con["concordium_mainnet"]
            self.mainnet: Dict[Collections, AsyncIOMotorCollection] = {}  # type: ignore
            for collection in Collections:
                self.mainnet[collection] = self.mainnet_db[collection.value]

            self.testnet_db = con["concordium_testnet"]
            self.testnet: Dict[Collections, AsyncIOMotorCollection] = {}  # type: ignore
            for collection in Collections:
                self.testnet[collection] = self.testnet_db[collection.value]

            # console.log(f'Motor: {con.server_info()["version"]}')
        except Exception as e:
            print(e)
            tooter.send(
                channel=TooterChannel.NOTIFIER,
                message=f"BOT ERROR! Cannot connect to Motor MongoDB, with error: {e}",
                notifier_type=TooterType.MONGODB_ERROR,
            )



from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *  # noqa: F403


class CredentialElement(Enum):
    firstName = "First Name"
    lastName = "Last Name"
    sex = "Sex"
    dob = "Date of Birth"
    countryOfResidence = "Country of Residence"
    nationality = "Nationatility"
    idDocType = "Identity Document Type"
    idDocNo = "Identity Document Number"
    idDocIssuer = "Identity Document Issuer"
    idDocIssuedAt = "ID Valid from"
    idDocExpiresAt = "ID Valid to"
    nationalIdNo = "National ID number"
    taxIdNo = "Tax ID number"


class CredentialDocType(Enum):
    na = "0"
    Passport = "1"
    National_ID_Card = "2"
    Driving_License = "3"
    Immigration_Card = "4"


class Credentials:
    """
    This class processes credential information as retrieved from the node.
    """

    def __init__(self):
        pass

    def determine_id_providers(self, ac):
        """
        Input to this method is the output from the node.
        """
        credentials = []
        for key, v in ac.items():
            v: CCD_AccountCredential  # noqa: F405
            if v.initial:
                v = v.initial
            elif v.normal:
                v = v.normal

            c = {
                "ip_identity": v.ip_id,
                "created_at": v.policy.created_at,
                "valid_to": v.policy.valid_to,
            }
            if len(v.policy.attributes.keys()) > 0:
                for key, revealedAttribute in v.policy.attributes.items():
                    value = revealedAttribute
                    c.update({CredentialElement[key].value: value})
            credentials.append(c)
        return credentials


class Identity:
    def __init__(self, account_info: CCD_AccountInfo):  # noqa: F405
        if account_info:
            self.credentials = Credentials().determine_id_providers(
                account_info.credentials
            )
            self.threshold = account_info.threshold
        else:
            self.credentials = []
            self.threshold = 0



# ruff: noqa: F403, F405, E402
from pydantic import BaseModel
import datetime as dt
from enum import Enum
from typing import Optional

# this leads to circluar import!
# from ccdexplorer_fundamentals.cis import MongoTypeLoggedEvent


################
#
# CCD_ is the prefix for all Pydantic classes
# to prevent namespace collision with the
# protobuf generated classes.
#
################


class CredentialElement(Enum):
    firstName = 0
    lastName = 1
    sex = 2
    dob = 3
    countryOfResidence = 4
    nationality = 5
    idDocType = 6
    idDocNo = 7
    idDocIssuer = 8
    idDocIssuedAt = 9
    idDocExpiresAt = 10
    nationalIdNo = 11
    taxIdNo = 12


class CredentialDocType(Enum):
    na = "0"
    Passport = "1"
    National_ID_Card = "2"
    Driving_License = "3"
    Immigration_Card = "4"


class ProtocolVersions(Enum):
    PROTOCOL_VERSION_1 = 0
    PROTOCOL_VERSION_2 = 1
    PROTOCOL_VERSION_3 = 2
    PROTOCOL_VERSION_4 = 3
    PROTOCOL_VERSION_5 = 4
    PROTOCOL_VERSION_6 = 5


class OpenStatus(Enum):
    open_for_all = 0
    closed_for_new = 1
    closed_for_all = 2


CCD_ArIdentity = int
CCD_IpIdentity = int
CCD_ArPublicKey = str
CCD_IpVerifyKey = str
CCD_IpCdiVerifyKey = str
CCD_BlockHash = str
CCD_TransactionHash = str
CCD_AccountAddress = str
CCD_AccountIndex = int
CCD_DelegatorId = CCD_AccountIndex
microCCD = int
CCD_BakerId = int
CCD_Epoch = int
CCD_Energy = int
CCD_DurationSeconds = int
CCD_Duration = int
CCD_ModuleRef = str
CCD_ContractEvent = str
CCD_Memo = str
CCD_RegisteredData = str
CCD_BakerSignatureVerifyKey = str
CCD_OpenStatus = int
CCD_BakerElectionVerifyKey = str
CCD_BakerAggregationVerifyKey = str
CCD_Parameter = str
CCD_AmountFraction = float
CCD_ReceiveName = str
CCD_CredentialsPerBlockLimit = int
CCD_LeadershipElectionNonce = str
CCD_CredentialRegistrationId = str
CCD_Sha256Hash = str
CCD_ElectionDifficulty = float
CCD_UpdatePublicKey = str
CCD_UpdateKeysIndex = int
CCD_UpdateKeysThreshold = int
CCD_TransactionTime = int
CCD_SignatureThreshold = int
CCD_IdentityProviderIdentity = int
CCD_ArThreshold = int
CCD_Commitment = str
CCD_VersionedModuleSource_ModuleSourceV0 = str
CCD_VersionedModuleSource_ModuleSourceV1 = str
CCD_ProtocolVersion = int
CCD_StateHash = str
CCD_TimeStamp = dt.datetime

CCD_SequenceNumber = int
CCD_StakePendingChange_Remove = CCD_TimeStamp
CCD_ContractStateV0 = str
CCD_InitName = str
CCD_EncryptedAmount = str
CCD_Empty = None
CCD_AccountThreshold = int
CCD_Policy_Attributes = str
CCD_Round = int
CCD_Epoch = int

# class CCD_Policy_Attributes(BaseModel):
#     key: int
#     value: str


class CCD_ContractAddress(BaseModel):
    index: int
    subindex: int

    def to_str(self):
        return f"<{self.index},{self.subindex}>"

    @classmethod
    def from_str(self, str_repr: str):
        """
        Returns a CCD_ContractAddress when the input is formed as '<2350,0>'.
        """
        contract_string = str_repr.split("-")[0]
        c = CCD_ContractAddress(
            **{
                "index": contract_string.split(",")[0][1:],
                "subindex": contract_string.split(",")[1][:-1],
            }
        )

        return c

    @classmethod
    def from_index(self, index: int, subindex: int):
        s = CCD_ContractAddress(**{"index": index, "subindex": subindex})
        return s


class CCD_Address(BaseModel):
    account: Optional[CCD_AccountAddress] = None
    contract: Optional[CCD_ContractAddress] = None


class CCD_RejectReason_InvalidInitMethod(BaseModel):
    module_ref: CCD_ModuleRef
    init_name: CCD_InitName


class CCD_RejectReason_InvalidReceiveMethod(BaseModel):
    module_ref: CCD_ModuleRef
    receive_name: CCD_ReceiveName


class CCD_RejectReason_AmountTooLarge(BaseModel):
    address: CCD_Address
    amount: microCCD


class CCD_RejectReason_RejectedInit(BaseModel):
    reject_reason: int


class CCD_RejectReason_RejectedReceive(BaseModel):
    reject_reason: int
    contract_address: CCD_ContractAddress
    receive_name: CCD_ReceiveName
    parameter: CCD_Parameter


class CCD_RejectReason_DuplicateCredIds(BaseModel):
    ids: list[CCD_CredentialRegistrationId]


class CCD_RejectReason_NonExistentCredIds(BaseModel):
    ids: list[CCD_CredentialRegistrationId]


class CCD_RejectReason(BaseModel):
    module_not_wf: Optional[CCD_Empty] = None
    module_hash_already_exists: Optional[CCD_ModuleRef] = None
    invalid_account_reference: Optional[CCD_AccountAddress] = None
    invalid_init_method: Optional[CCD_RejectReason_InvalidInitMethod] = None
    invalid_receive_method: Optional[CCD_RejectReason_InvalidReceiveMethod] = None
    invalid_module_reference: Optional[CCD_ModuleRef] = None
    invalid_contract_address: Optional[CCD_ContractAddress] = None
    runtime_failure: Optional[CCD_Empty] = None
    amount_too_large: Optional[CCD_RejectReason_AmountTooLarge] = None
    serialization_failure: Optional[CCD_Empty] = None
    out_of_energy: Optional[CCD_Empty] = None
    rejected_init: Optional[CCD_RejectReason_RejectedInit] = None
    rejected_receive: Optional[CCD_RejectReason_RejectedReceive] = None
    invalid_proof: Optional[CCD_Empty] = None
    already_a_baker: Optional[CCD_BakerId] = None
    not_a_baker: Optional[CCD_AccountAddress] = None
    insufficient_balance_for_baker_stake: Optional[CCD_Empty] = None
    stake_under_minimum_threshold_for_baking: Optional[CCD_Empty] = None
    baker_in_cooldown: Optional[CCD_Empty] = None
    duplicate_aggregation_key: Optional[CCD_BakerAggregationVerifyKey] = None
    non_existent_credential_id: Optional[CCD_Empty] = None
    key_index_already_in_use: Optional[CCD_Empty] = None
    invalid_account_threshold: Optional[CCD_Empty] = None
    invalid_credential_key_sign_threshold: Optional[CCD_Empty] = None
    invalid_encrypted_amount_transfer_proof: Optional[CCD_Empty] = None
    invalid_transfer_to_public_proof: Optional[CCD_Empty] = None
    encrypted_amount_self_transfer: Optional[CCD_AccountAddress] = None
    invalid_index_on_encrypted_transfer: Optional[CCD_Empty] = None
    zero_scheduledAmount: Optional[CCD_Empty] = None
    non_increasing_schedule: Optional[CCD_Empty] = None
    first_scheduled_release_expired: Optional[CCD_Empty] = None
    scheduled_self_transfer: Optional[CCD_AccountAddress] = None
    invalid_credentials: Optional[CCD_Empty] = None
    duplicate_cred_ids: Optional[CCD_RejectReason_DuplicateCredIds] = None
    non_existent_cred_ids: Optional[CCD_RejectReason_NonExistentCredIds] = None
    remove_first_credential: Optional[CCD_Empty] = None
    credential_holder_did_not_sign: Optional[CCD_Empty] = None
    not_allowed_multiple_credentials: Optional[CCD_Empty] = None
    not_allowed_to_receive_encrypted: Optional[CCD_Empty] = None
    not_allowed_to_handle_encrypted: Optional[CCD_Empty] = None
    missing_baker_add_parameters: Optional[CCD_Empty] = None
    finalization_reward_commission_not_in_range: Optional[CCD_Empty] = None
    baking_reward_commission_not_in_range: Optional[CCD_Empty] = None
    transaction_fee_commission_not_in_range: Optional[CCD_Empty] = None
    already_a_delegator: Optional[CCD_Empty] = None
    insufficient_balance_for_delegation_stake: Optional[CCD_Empty] = None
    missing_delegation_add_parameters: Optional[CCD_Empty] = None
    insufficient_delegation_stake: Optional[CCD_Empty] = None
    delegator_in_cooldown: Optional[CCD_Empty] = None
    not_a_delegator: Optional[CCD_AccountAddress] = None
    delegation_target_not_a_baker: Optional[CCD_BakerId] = None
    stake_over_maximum_threshold_for_pool: Optional[CCD_Empty] = None
    pool_would_become_over_delegated: Optional[CCD_Empty] = None
    pool_closed: Optional[CCD_Empty] = None


class CCD_CredentialType(Enum):
    initial = 0
    normal = 1


class CCD_StakePendingChange_Reduce(BaseModel):
    new_stake: microCCD
    effective_time: CCD_TimeStamp


class CCD_StakePendingChange(BaseModel):
    reduce: Optional[CCD_StakePendingChange_Reduce] = None
    remove: Optional[CCD_StakePendingChange_Remove] = None


class CCD_BakerStakePendingChange_Reduce(BaseModel):
    reduced_equity_capital: microCCD
    effective_time: CCD_TimeStamp


class CCD_BakerStakePendingChange_Remove(BaseModel):
    effective_time: CCD_TimeStamp


class CCD_BakerStakePendingChange(BaseModel):
    reduce: Optional[CCD_BakerStakePendingChange_Reduce] = None
    remove: Optional[CCD_BakerStakePendingChange_Remove] = None


class CCD_DelegatorInfo(BaseModel):
    account: CCD_AccountAddress
    stake: microCCD
    pending_change: Optional[CCD_StakePendingChange] = None


class CCD_DelegatorRewardPeriodInfo(BaseModel):
    account: CCD_AccountAddress
    stake: microCCD


class CCD_CurrentPaydayStatus(BaseModel):
    baker_equity_capital: microCCD
    blocks_baked: int
    delegated_capital: microCCD
    effective_stake: microCCD
    finalization_live: bool
    lottery_power: float
    transaction_fees_earned: microCCD


class CCD_CommissionRates(BaseModel):
    baking: float
    finalization: float
    transaction: float


class CCD_BakerPoolInfo(BaseModel):
    commission_rates: CCD_CommissionRates
    url: str
    open_status: str


class CCD_PoolInfo(BaseModel):
    all_pool_total_capital: microCCD
    address: CCD_AccountAddress
    equity_capital: microCCD
    baker: int
    equity_pending_change: Optional[CCD_BakerStakePendingChange] = None
    current_payday_info: Optional[CCD_CurrentPaydayStatus] = None
    delegated_capital: microCCD
    delegated_capital_cap: microCCD
    pool_info: CCD_BakerPoolInfo
    # poolType: str = None


class CCD_PassiveDelegationInfo(BaseModel):
    all_pool_total_capital: microCCD
    delegated_capital: microCCD
    current_payday_transaction_fees_earned: microCCD
    current_payday_delegated_capital: microCCD
    commission_rates: CCD_CommissionRates


class CCD_FinalizedBlockInfo(BaseModel):
    hash: CCD_BlockHash
    height: int


class CCD_BlockInfo(BaseModel):
    arrive_time: Optional[CCD_TimeStamp] = None
    baker: Optional[int] = None
    hash: CCD_BlockHash
    height: int
    last_finalized_block: CCD_BlockHash
    parent_block: CCD_BlockHash
    receive_time: Optional[CCD_TimeStamp] = None
    slot_number: Optional[int] = None
    slot_time: CCD_TimeStamp
    era_block_height: int
    finalized: bool
    genesis_index: int
    transaction_count: int
    transactions_energy_cost: int
    transactions_size: int
    transaction_hashes: Optional[list[CCD_TransactionHash]] = None
    state_hash: Optional[CCD_StateHash] = None
    protocol_version: Optional[str] = (
        None  # note this is not optional from the specification,
    )
    # but this type is also used in retrieving blockInfo from MongoDB, where protocol_version
    # isn't stored for all blocks (currently).
    round: Optional[CCD_Round] = None
    epoch: Optional[CCD_Epoch] = None


class CCD_AccountTransactionEffects_None(BaseModel):
    transaction_type: Optional[int] = None
    reject_reason: CCD_RejectReason


class CCD_ContractInitializedEvent(BaseModel):
    contract_version: int
    origin_ref: CCD_ModuleRef
    address: CCD_ContractAddress
    amount: microCCD = 0
    init_name: str
    events: list[CCD_ContractEvent]


class CCD_InstanceUpdatedEvent(BaseModel):
    contract_version: int
    address: CCD_ContractAddress
    instigator: CCD_Address
    amount: microCCD
    parameter: CCD_Parameter
    receive_name: CCD_ReceiveName
    events: Optional[list[CCD_ContractEvent]] = None


class CCD_ContractTraceElement_Interrupted(BaseModel):
    address: CCD_ContractAddress
    events: list[CCD_ContractEvent]


class CCD_ContractTraceElement_Resumed(BaseModel):
    address: CCD_ContractAddress
    success: bool


class CCD_ContractTraceElement_Transferred(BaseModel):
    sender: CCD_ContractAddress
    amount: microCCD
    receiver: CCD_AccountAddress


class CCD_ContractTraceElement(BaseModel):
    updated: Optional[CCD_InstanceUpdatedEvent] = None
    transferred: Optional[CCD_ContractTraceElement_Transferred] = None
    interrupted: Optional[CCD_ContractTraceElement_Interrupted] = None
    resumed: Optional[CCD_ContractTraceElement_Resumed] = None


class CCD_ContractUpdateIssued(BaseModel):
    effects: list[CCD_ContractTraceElement]


class CCD_EncryptedAmountRemovedEvent(BaseModel):
    account: CCD_AccountAddress
    new_amount: CCD_EncryptedAmount
    input_amount: CCD_EncryptedAmount
    up_to_index: int


class CCD_NewEncryptedAmountEvent(BaseModel):
    receiver: CCD_AccountAddress
    new_index: int
    encrypted_amount: CCD_EncryptedAmount


class CCD_AccountTransactionEffects_EncryptedAmountTransferred(BaseModel):
    removed: Optional[CCD_EncryptedAmountRemovedEvent] = None
    added: Optional[CCD_NewEncryptedAmountEvent] = None
    memo: Optional[CCD_Memo] = None


class CCD_EncryptedSelfAmountAddedEvent(BaseModel):
    account: CCD_AccountAddress
    new_amount: CCD_EncryptedAmount
    amount: microCCD


class CCD_AccountTransactionEffects_TransferredToPublic(BaseModel):
    removed: CCD_EncryptedAmountRemovedEvent
    amount: microCCD


class CCD_AccountTransactionEffects_CredentialsUpdated(BaseModel):
    new_cred_ids: list[CCD_CredentialRegistrationId]
    removed_cred_ids: list[CCD_CredentialRegistrationId]
    new_threshold: CCD_AccountThreshold


class CCD_BakerKeysEvent(BaseModel):
    baker_id: CCD_BakerId
    account: CCD_AccountAddress
    sign_key: CCD_BakerSignatureVerifyKey
    election_key: CCD_BakerElectionVerifyKey
    aggregation_key: CCD_BakerAggregationVerifyKey


class CCD_BakerAdded(BaseModel):
    keys_event: CCD_BakerKeysEvent
    stake: microCCD
    restake_earnings: bool


class CCD_AccountTransfer(BaseModel):
    amount: microCCD = 0
    receiver: CCD_AccountAddress
    memo: Optional[CCD_Memo] = None


class CCD_NewRelease(BaseModel):
    timestamp: CCD_TimeStamp
    amount: microCCD


class CCD_TransferredWithSchedule(BaseModel):
    receiver: CCD_AccountAddress
    amount: list[CCD_NewRelease]
    memo: Optional[CCD_Memo] = None


class CCD_BakerStakeUpdatedData(BaseModel):
    baker_id: CCD_BakerId
    new_stake: microCCD
    increased: bool


class CCD_BakerStakeUpdated(BaseModel):
    update: CCD_BakerStakeUpdatedData


class CCD_BakerStakeIncreased(BaseModel):
    baker_id: CCD_BakerId
    new_stake: microCCD


class CCD_BakerStakeDecreased(BaseModel):
    baker_id: CCD_BakerId
    new_stake: microCCD


class CCD_BakerRestakeEarningsUpdated(BaseModel):
    baker_id: CCD_BakerId
    restake_earnings: bool


class CCD_BakerSetOpenStatus(BaseModel):
    baker_id: CCD_BakerId
    open_status: CCD_OpenStatus


class CCD_BakerSetMetadataUrl(BaseModel):
    baker_id: CCD_BakerId
    url: str


class CCD_BakerSetTransactionFeeCommission(BaseModel):
    baker_id: CCD_BakerId
    transaction_fee_commission: CCD_AmountFraction


class CCD_BakerSetBakingRewardCommission(BaseModel):
    baker_id: CCD_BakerId
    baking_reward_commission: CCD_AmountFraction


class CCD_BakerSetFinalizationRewardCommission(BaseModel):
    baker_id: CCD_BakerId
    finalization_reward_commission: CCD_AmountFraction


class CCD_BakerEvent(BaseModel):
    baker_added: Optional[CCD_BakerAdded] = None
    baker_removed: Optional[CCD_BakerId] = None
    baker_stake_increased: Optional[CCD_BakerStakeIncreased] = None
    baker_stake_decreased: Optional[CCD_BakerStakeDecreased] = None
    baker_restake_earnings_updated: Optional[CCD_BakerRestakeEarningsUpdated] = None
    baker_keys_updated: Optional[CCD_BakerKeysEvent] = None
    baker_set_open_status: Optional[CCD_BakerSetOpenStatus] = None
    baker_set_metadata_url: Optional[CCD_BakerSetMetadataUrl] = None
    baker_set_transaction_fee_commission: Optional[
        CCD_BakerSetTransactionFeeCommission
    ] = None
    baker_set_baking_reward_commission: Optional[CCD_BakerSetBakingRewardCommission] = (
        None
    )
    baker_set_finalization_reward_commission: Optional[
        CCD_BakerSetFinalizationRewardCommission
    ] = None


class CCD_BakerConfigured(BaseModel):
    events: list[CCD_BakerEvent]


class CCD_DelegationStakeIncreased(BaseModel):
    delegator_id: CCD_DelegatorId
    new_stake: microCCD


class CCD_DelegationStakeDecreased(BaseModel):
    delegator_id: CCD_DelegatorId
    new_stake: microCCD


class CCD_DelegationSetRestakeEarnings(BaseModel):
    delegator_id: CCD_DelegatorId
    restake_earnings: bool


class CCD_DelegationTarget(BaseModel):
    passive_delegation: Optional[bool] = None
    baker: Optional[CCD_BakerId] = None


class CCD_DelegationSetDelegationTarget(BaseModel):
    delegator_id: CCD_DelegatorId
    delegation_target: CCD_DelegationTarget


class CCD_DelegationEvent(BaseModel):
    delegation_added: Optional[CCD_DelegatorId] = None
    delegation_removed: Optional[CCD_DelegatorId] = None
    delegation_stake_increased: Optional[CCD_DelegationStakeIncreased] = None
    delegation_stake_decreased: Optional[CCD_DelegationStakeDecreased] = None
    delegation_set_restake_earnings: Optional[CCD_DelegationSetRestakeEarnings] = None
    delegation_set_delegation_target: Optional[CCD_DelegationSetDelegationTarget] = None


class CCD_DelegationConfigured(BaseModel):
    events: list[CCD_DelegationEvent]


class CCD_BakerEvent_BakerRestakeEarningsUpdated(BaseModel):
    baker_id: CCD_BakerId
    restake_earnings: bool


class CCD_AccountTransactionEffects(BaseModel):
    none: Optional[CCD_AccountTransactionEffects_None] = None
    module_deployed: Optional[CCD_ModuleRef] = None
    contract_initialized: Optional[CCD_ContractInitializedEvent] = None
    contract_update_issued: Optional[CCD_ContractUpdateIssued] = None
    account_transfer: Optional[CCD_AccountTransfer] = None
    baker_added: Optional[CCD_BakerAdded] = None
    baker_removed: Optional[CCD_BakerId] = None
    baker_stake_updated: Optional[CCD_BakerStakeUpdated] = None
    baker_restake_earnings_updated: Optional[
        CCD_BakerEvent_BakerRestakeEarningsUpdated
    ] = None
    baker_keys_updated: Optional[CCD_BakerKeysEvent] = None
    encrypted_amount_transferred: Optional[
        CCD_AccountTransactionEffects_EncryptedAmountTransferred
    ] = None
    transferred_to_encrypted: CCD_EncryptedSelfAmountAddedEvent = None
    transferred_to_public: CCD_AccountTransactionEffects_TransferredToPublic = None
    transferred_with_schedule: CCD_TransferredWithSchedule = None
    credential_keys_updated: CCD_CredentialRegistrationId = None
    credentials_updated: CCD_AccountTransactionEffects_CredentialsUpdated = None
    data_registered: CCD_RegisteredData = None
    baker_configured: CCD_BakerConfigured = None
    delegation_configured: CCD_DelegationConfigured = None


class CCD_AccountTransactionDetails(BaseModel):
    cost: microCCD
    sender: CCD_AccountAddress
    outcome: str
    effects: CCD_AccountTransactionEffects


class CCD_AccountCreationDetails(BaseModel):
    credential_type: int
    address: CCD_AccountAddress
    reg_id: CCD_CredentialRegistrationId


class CCD_ProtocolUpdate(BaseModel):
    message_: str
    specification_url: str
    specificationHash: CCD_Sha256Hash
    specification_auxiliary_data: Optional[bytes] = None


class CCD_Ratio(BaseModel):
    # note these are ints, but need to convert to str
    # as the int amounts are sometimes too big (for MongoDB)
    numerator: str
    denominator: str


class CCD_ExchangeRate(BaseModel):
    numerator: str
    denominator: str


class CCD_MintRate(BaseModel):
    mantissa: int
    exponent: int


class CCD_MintDistributionCpv0(BaseModel):
    mint_per_slot: CCD_MintRate
    baking_reward: CCD_AmountFraction
    finalization_reward: CCD_AmountFraction


class CCD_TransactionFeeDistribution(BaseModel):
    baker: Optional[CCD_AmountFraction] = None
    gas_account: Optional[CCD_AmountFraction] = None


class CCD_GasRewards(BaseModel):
    baker: CCD_AmountFraction
    finalization_proof: CCD_AmountFraction
    account_creation: CCD_AmountFraction
    chain_update: CCD_AmountFraction


class CCD_HigherLevelKeys(BaseModel):
    keys: list[CCD_UpdatePublicKey]
    threshold: CCD_UpdateKeysThreshold


class CCD_AccessStructure(BaseModel):
    access_public_keys: list[CCD_UpdateKeysIndex]
    access_threshold: CCD_UpdateKeysThreshold


class CCD_AuthorizationsV0(BaseModel):
    keys: list[CCD_UpdatePublicKey]
    emergency: CCD_AccessStructure
    protocol: CCD_AccessStructure
    parameter_election_difficulty: Optional[CCD_AccessStructure] = None
    parameter_euro_per_energy: CCD_AccessStructure
    parameter_micro_CCD_per_euro: CCD_AccessStructure
    parameter_foundation_account: CCD_AccessStructure
    parameter_mint_distribution: CCD_AccessStructure
    parameter_transaction_fee_distribution: CCD_AccessStructure
    parameter_gas_rewards: CCD_AccessStructure
    pool_parameters: CCD_AccessStructure
    add_anonymity_revoker: CCD_AccessStructure
    add_identity_provider: CCD_AccessStructure


class CCD_AuthorizationsV1(BaseModel):
    v0: CCD_AuthorizationsV0
    parameter_cooldown: CCD_AccessStructure
    parameter_time: CCD_AccessStructure


class CCD_RootUpdate(BaseModel):
    root_keys_update: CCD_HigherLevelKeys
    level_1_keys_update: CCD_HigherLevelKeys
    level_2_keys_update_v0: CCD_AuthorizationsV0
    level_2_keys_update_v1: CCD_AuthorizationsV1


class CCD_Level1Update(BaseModel):
    level_1_keys_update: Optional[CCD_HigherLevelKeys] = None
    level_2_keys_update_v0: Optional[CCD_AuthorizationsV0] = None
    level_2_keys_update_v1: Optional[CCD_AuthorizationsV1] = None


class CCD_Description(BaseModel):
    name: str
    url: str
    description: str


class CCD_ArInfo(BaseModel):
    identity: CCD_ArIdentity
    description: CCD_Description
    public_key: CCD_ArPublicKey


class CCD_IpInfo(BaseModel):
    identity: CCD_IpIdentity
    description: CCD_Description
    verify_key: CCD_IpVerifyKey
    cdi_verify_key: CCD_IpCdiVerifyKey


class CCD_CooldownParametersCpv1(BaseModel):
    pool_owner_cooldown: Optional[CCD_DurationSeconds] = None
    delegator_cooldown: Optional[CCD_DurationSeconds] = None


class CCD_InclusiveRangeAmountFraction(BaseModel):
    min: CCD_AmountFraction
    max_: CCD_AmountFraction


class CCD_CommissionRanges(BaseModel):
    finalization: CCD_InclusiveRangeAmountFraction
    baking: CCD_InclusiveRangeAmountFraction
    transaction: CCD_InclusiveRangeAmountFraction


class CCD_CapitalBound(BaseModel):
    value: CCD_AmountFraction


class CCD_LeverageFactor(BaseModel):
    value: CCD_Ratio


class CCD_PoolParametersCpv1(BaseModel):
    passive_finalization_commission: CCD_AmountFraction
    passive_baking_commission: CCD_AmountFraction
    passive_transaction_commission: CCD_AmountFraction
    commission_bounds: CCD_CommissionRanges
    minimum_equity_capital: microCCD
    capital_bound: CCD_CapitalBound
    leverage_bound: CCD_LeverageFactor


class CCD_RewardPeriodLength(BaseModel):
    value: CCD_Epoch


class CCD_TimeParametersCpv1(BaseModel):
    reward_period_length: CCD_Epoch
    mint_per_payday: CCD_MintRate


class CCD_MintDistributionCpv1(BaseModel):
    baking_reward: CCD_AmountFraction
    finalization_reward: CCD_AmountFraction


class CCD_BakerStakeThreshold(BaseModel):
    baker_stake_threshold: microCCD


class CCD_TransactionType(BaseModel):
    type: str
    contents: str


class CCD_FinalizationCommitteeParameters(BaseModel):
    minimum_finalizers: int
    maximum_finalizers: int
    finalizer_relative_stake_threshold: CCD_AmountFraction


class CCD_UpdatePayload(BaseModel):
    protocol_update: Optional[CCD_ProtocolUpdate] = None
    election_difficulty_update: Optional[CCD_ElectionDifficulty] = None
    euro_per_energy_update: Optional[CCD_ExchangeRate] = None
    micro_ccd_per_euro_update: Optional[CCD_ExchangeRate] = None
    foundation_account_update: Optional[CCD_AccountAddress] = None
    mint_distribution_update: Optional[CCD_MintDistributionCpv0] = None
    transaction_fee_distribution_update: Optional[CCD_TransactionFeeDistribution] = None
    baker_stake_threshold_update: Optional[CCD_BakerStakeThreshold] = None
    root_update: Optional[CCD_RootUpdate] = None
    level_1_update: Optional[CCD_Level1Update] = None
    add_anonymity_revoker_update: Optional[CCD_ArInfo] = None
    add_identity_provider_update: Optional[CCD_IpInfo] = None
    cooldown_parameters_cpv_1_update: Optional[CCD_CooldownParametersCpv1] = None
    pool_parameters_cpv_1_update: Optional[CCD_PoolParametersCpv1] = None
    time_parameters_cpv_1_update: Optional[CCD_TimeParametersCpv1] = None
    mint_distribution_cpv_1_update: Optional[CCD_MintDistributionCpv1] = None
    finalization_committee_parameters_update: Optional[
        CCD_FinalizationCommitteeParameters
    ] = None


class CCD_UpdateDetails(BaseModel):
    effective_time: CCD_TransactionTime = None
    payload: CCD_UpdatePayload


class CCD_ShortBlockInfo(BaseModel):
    height: int
    hash: CCD_BlockHash
    slot_time: CCD_TimeStamp


class CCD_BlockItemSummary(BaseModel):
    index: int = 0
    energy_cost: int = None
    hash: CCD_TransactionHash
    type: CCD_TransactionType = None
    account_transaction: CCD_AccountTransactionDetails = None
    account_creation: CCD_AccountCreationDetails = None
    update: CCD_UpdateDetails = None
    block_info: CCD_ShortBlockInfo = None


class CCD_Block(BaseModel):
    transaction_summaries: list[CCD_BlockItemSummary]


class CCD_Release(BaseModel):
    timestamp: CCD_TimeStamp
    amount: microCCD
    transactions: list[CCD_TransactionHash]


class CCD_ReleaseSchedule(BaseModel):
    schedules: list[CCD_Release]
    total: int


class CCD_BakerInfo(BaseModel):
    aggregation_key: str
    election_key: str
    baker_id: int
    signature_key: str


class CCD_AccountStakingInfo_Baker(BaseModel):
    baker_info: CCD_BakerInfo
    pool_info: CCD_BakerPoolInfo
    pending_change: CCD_StakePendingChange
    restake_earnings: bool
    staked_amount: microCCD


# class CCD_DelegationTarget(BaseModel):
#     baker_id: int = None
#     passive: str = None


class CCD_AccountStakingInfo_Delegator(BaseModel):
    target: CCD_DelegationTarget
    pending_change: CCD_StakePendingChange
    restake_earnings: bool
    staked_amount: microCCD


class CCD_AccountStakingInfo(BaseModel):
    baker: Optional[CCD_AccountStakingInfo_Baker] = None
    delegator: Optional[CCD_AccountStakingInfo_Delegator] = None


class CCD_AccountVerifyKey(BaseModel):
    ed25519_key: str


# class CCD_KeysEntry(BaseModel):
#     key: int
#     value: CCD_AccountVerifyKey


class CCD_CredentialPublicKeys(BaseModel):
    keys: dict[str, CCD_AccountVerifyKey]
    threshold: CCD_SignatureThreshold


class CCD_YearMonth(BaseModel):
    year: int
    month: int


class CCD_Policy(BaseModel):
    created_at: CCD_YearMonth
    valid_to: CCD_YearMonth
    attributes: dict[str, CCD_Policy_Attributes]


class CCD_InitialCredentialValues(BaseModel):
    credential_public_keys: CCD_CredentialPublicKeys
    cred_id: CCD_CredentialRegistrationId
    ip_id: CCD_IdentityProviderIdentity
    policy: CCD_Policy


class CCD_ChainArData(BaseModel):
    enc_id_cred_pub_share: str


# class CCD_ArDataEntry(BaseModel):
#     key: int
#     value: CCD_ChainArData


class CCD_CredentialCommitments_AttributesEntry(BaseModel):
    key: int
    value: CCD_Commitment


class CCD_CredentialCommitments(BaseModel):
    prf: CCD_Commitment
    cred_counter: CCD_Commitment
    max_accounts: CCD_Commitment
    attributes: dict[str, CCD_Commitment]
    id_cred_sec_sharing_coeff: list[CCD_Commitment]


class CCD_NormalCredentialValues(BaseModel):
    credential_public_keys: CCD_CredentialPublicKeys
    cred_id: CCD_CredentialRegistrationId
    ip_id: CCD_IdentityProviderIdentity
    policy: CCD_Policy
    ar_threshold: CCD_ArThreshold
    ar_data: dict[str, CCD_ChainArData]
    commitments: CCD_CredentialCommitments


class CCD_AccountCredential(BaseModel):
    initial: CCD_InitialCredentialValues = None
    normal: CCD_NormalCredentialValues = None


class CCD_EncryptedBalance(BaseModel):
    self_amount: CCD_EncryptedAmount
    start_index: int
    aggregated_amount: CCD_EncryptedAmount = None
    num_aggregated: int = None
    incoming_amounts: list[CCD_EncryptedAmount]


class CCD_AccountInfo(BaseModel):
    address: str
    amount: microCCD
    stake: Optional[CCD_AccountStakingInfo] = None
    credentials: dict[str, CCD_AccountCredential]
    encrypted_balance: CCD_EncryptedBalance
    encryption_key: str
    index: int
    schedule: Optional[CCD_ReleaseSchedule] = None
    threshold: int
    sequence_number: CCD_SequenceNumber


class CCD_TokenomicsInfo_V0(BaseModel):
    total_amount: microCCD
    total_encrypted_amount: microCCD
    baking_reward_account: microCCD
    finalization_reward_account: microCCD
    gas_account: microCCD
    protocol_version: int


class CCD_TokenomicsInfo_V1(BaseModel):
    total_amount: microCCD
    total_encrypted_amount: microCCD
    baking_reward_account: microCCD
    finalization_reward_account: microCCD
    gas_account: microCCD
    foundation_transaction_rewards: microCCD
    next_payday_time: CCD_TimeStamp
    next_payday_mint_rate: CCD_MintRate
    total_staked_capital: microCCD
    protocol_version: int


class CCD_TokenomicsInfo(BaseModel):
    v0: Optional[CCD_TokenomicsInfo_V0] = None
    v1: Optional[CCD_TokenomicsInfo_V1] = None


class CCD_InstanceInfo_V0(BaseModel):
    model: CCD_ContractStateV0
    owner: CCD_AccountAddress
    amount: microCCD
    methods: list[CCD_ReceiveName]
    name: CCD_InitName
    source_module: CCD_ModuleRef


class CCD_InstanceInfo_V1(BaseModel):
    owner: CCD_AccountAddress
    amount: microCCD
    methods: list[CCD_ReceiveName]
    name: CCD_InitName
    source_module: CCD_ModuleRef


class CCD_InstanceInfo(BaseModel):
    v0: CCD_InstanceInfo_V0
    v1: CCD_InstanceInfo_V1


class CCD_BlocksAtHeightResponse(BaseModel):
    blocks: list[CCD_BlockHash]


class CCD_BlockSpecialEvent_PaydayPoolReward(BaseModel):
    pool_owner: CCD_BakerId = None
    transaction_fees: microCCD
    baker_reward: microCCD
    finalization_reward: microCCD


class CCD_BlockSpecialEvent_BlockAccrueReward(BaseModel):
    transaction_fees: microCCD
    old_gas_account: microCCD
    new_gas_account: microCCD
    baker_reward: microCCD
    passive_reward: microCCD
    foundation_charge: microCCD
    baker: CCD_BakerId


class CCD_BlockSpecialEvent_PaydayAccountReward(BaseModel):
    account: CCD_AccountAddress
    transaction_fees: microCCD
    baker_reward: microCCD
    finalization_reward: microCCD


class CCD_BlockSpecialEvent_PaydayFoundationReward(BaseModel):
    foundation_account: CCD_AccountAddress
    development_charge: microCCD


class CCD_BlockSpecialEvent_BlockReward(BaseModel):
    transaction_fees: microCCD
    old_gas_account: microCCD
    new_gas_account: microCCD
    baker_reward: microCCD
    foundation_charge: microCCD
    foundation_account: CCD_AccountAddress
    baker: CCD_AccountAddress


class CCD_BlockSpecialEvent_AccountAmounts_Entry(BaseModel):
    account: CCD_AccountAddress
    amount: microCCD


class CCD_BlockSpecialEvent_AccountAmounts(BaseModel):
    entries: list[CCD_BlockSpecialEvent_AccountAmounts_Entry]


class CCD_BlockSpecialEvent_FinalizationRewards(BaseModel):
    finalization_rewards: CCD_BlockSpecialEvent_AccountAmounts
    remainder: microCCD


class CCD_BlockSpecialEvent_BakingRewards(BaseModel):
    baker_rewards: CCD_BlockSpecialEvent_AccountAmounts
    remainder: microCCD


class CCD_BlockSpecialEvent_Mint(BaseModel):
    mint_baking_reward: microCCD
    mint_finalization_reward: microCCD
    mint_platform_development_charge: microCCD
    foundation_account: CCD_AccountAddress


class CCD_BlockSpecialEvent(BaseModel):
    baking_rewards: CCD_BlockSpecialEvent_BakingRewards = None
    mint: CCD_BlockSpecialEvent_Mint = None
    finalization_rewards: CCD_BlockSpecialEvent_FinalizationRewards = None
    block_reward: CCD_BlockSpecialEvent_BlockReward = None
    payday_foundation_reward: CCD_BlockSpecialEvent_PaydayFoundationReward = None
    payday_account_reward: CCD_BlockSpecialEvent_PaydayAccountReward = None
    block_accrue_reward: CCD_BlockSpecialEvent_BlockAccrueReward = None
    payday_pool_reward: CCD_BlockSpecialEvent_PaydayPoolReward = None


class CCD_PendingUpdate(BaseModel):
    effective_time: CCD_TransactionTime
    root_keys: CCD_HigherLevelKeys = None
    level1_keys: CCD_HigherLevelKeys = None
    level2_keys_cpc_0: CCD_AuthorizationsV0 = None
    level2_keys_cpc_1: CCD_AuthorizationsV1 = None
    protocol: CCD_ProtocolUpdate = None
    election_difficulty: CCD_ElectionDifficulty = None
    euro_per_energy: CCD_ExchangeRate = None
    micro_ccd_per_euro: CCD_ExchangeRate = None
    foundation_account: CCD_AccountAddress = None
    mint_distribution_cpv_0: CCD_MintDistributionCpv0 = None
    mint_distribution_cpv_1: CCD_MintDistributionCpv1 = None
    transaction_fee_distribution: CCD_TransactionFeeDistribution = None
    gas_rewards: CCD_GasRewards
    pool_parameters_cpv_0: CCD_BakerStakeThreshold = None
    pool_parameters_cpv_0: CCD_PoolParametersCpv1 = None
    add_anonymity_revoker: CCD_ArInfo = None
    add_identity_provider: CCD_IpInfo = None
    cooldown_parameters: CCD_CooldownParametersCpv1 = None
    pool_parameters_cpv_1_update: CCD_PoolParametersCpv1 = None
    time_parameters: CCD_TimeParametersCpv1 = None


class CCD_ElectionInfo_Baker(BaseModel):
    baker: CCD_BakerId
    account: CCD_AccountAddress
    lottery_power: float


class CCD_ElectionInfo(BaseModel):
    election_difficulty: CCD_ElectionDifficulty = None
    election_nonce: CCD_LeadershipElectionNonce
    baker_election_info: list[CCD_ElectionInfo_Baker]


class CCD_ConsensusInfo(BaseModel):
    best_block: CCD_BlockHash
    genesis_block: CCD_BlockHash
    genesis_time: CCD_TimeStamp
    slot_duration: CCD_Duration = None
    epoch_duration: CCD_Duration
    last_finalized_block: CCD_BlockHash
    best_block_height: int
    last_finalized_block_height: int
    blocks_received_count: int
    block_last_received_time: CCD_TimeStamp = None
    block_receive_latency_ema: float
    block_receive_latency_emsd: float
    block_receive_period_ema: float = None
    block_receive_period_emsd: float = None
    blocks_verified_count: int
    block_last_arrived_time: CCD_TimeStamp = None
    block_arrive_latency_ema: float
    block_arrive_latency_emsd: float
    block_arrive_period_ema: float = None
    block_arrive_period_emsd: float = None
    transactions_per_block_ema: float
    transactions_per_block_emsd: float
    finalization_count: int
    last_finalized_time: CCD_TimeStamp = None
    finalization_period_ema: float = None
    finalization_period_emsd: float = None
    protocol_version: str
    genesis_index: int
    current_era_genesis_block: CCD_BlockHash
    current_era_genesis_time: CCD_TimeStamp
    current_timeout_duration: CCD_Duration = None
    current_round: CCD_Round = None
    current_epoch: CCD_Epoch = None
    trigger_block_time: CCD_TimeStamp = None


class CCD_VersionedModuleSource(BaseModel):
    v0: CCD_VersionedModuleSource_ModuleSourceV0 = None
    v1: CCD_VersionedModuleSource_ModuleSourceV1 = None


class CCD_ChainParametersV0(BaseModel):
    election_difficulty: CCD_ElectionDifficulty
    euro_per_energy: CCD_ExchangeRate
    micro_ccd_per_euro: CCD_ExchangeRate
    baker_cooldown_epochs: CCD_Epoch
    account_creation_limit: CCD_CredentialsPerBlockLimit
    mint_distribution: CCD_MintDistributionCpv0
    transaction_fee_distribution: CCD_TransactionFeeDistribution
    gas_rewards: CCD_GasRewards
    foundation_account: CCD_AccountAddress
    minimum_threshold_for_baking: microCCD
    root_keys: CCD_HigherLevelKeys
    level1_keys: CCD_HigherLevelKeys
    level2_keys: CCD_AuthorizationsV0


class CCD_ChainParametersV1(BaseModel):
    election_difficulty: CCD_ElectionDifficulty
    euro_per_energy: CCD_ExchangeRate
    micro_ccd_per_euro: CCD_ExchangeRate
    cooldown_parameters: CCD_CooldownParametersCpv1
    time_parameters: CCD_TimeParametersCpv1
    account_creation_limit: CCD_CredentialsPerBlockLimit
    mint_distribution: CCD_MintDistributionCpv1
    transaction_fee_distribution: CCD_TransactionFeeDistribution
    gas_rewards: CCD_GasRewards
    foundation_account: CCD_AccountAddress
    pool_parameters: CCD_PoolParametersCpv1
    root_keys: CCD_HigherLevelKeys
    level1_keys: CCD_HigherLevelKeys
    level2_keys: CCD_AuthorizationsV1


class CCD_TimeOutParameters(BaseModel):
    timeout_base: CCD_Duration
    timeout_increase: CCD_Ratio
    timeout_decrease: CCD_Ratio


class CCD_ConsensusParametersV1(BaseModel):
    timeout_parameters: CCD_TimeOutParameters
    min_block_time: CCD_Duration
    block_energy_limit: CCD_Energy


class CCD_GasRewardsV2(BaseModel):
    baker: CCD_AmountFraction
    account_creation: CCD_AmountFraction
    chain_update: CCD_AmountFraction


class CCD_ChainParametersV2(BaseModel):
    consensus_parameters: CCD_ConsensusParametersV1
    euro_per_energy: CCD_ExchangeRate
    micro_ccd_per_euro: CCD_ExchangeRate
    cooldown_parameters: CCD_CooldownParametersCpv1
    time_parameters: CCD_TimeParametersCpv1
    account_creation_limit: CCD_CredentialsPerBlockLimit
    mint_distribution: CCD_MintDistributionCpv1
    transaction_fee_distribution: CCD_TransactionFeeDistribution
    gas_rewards: CCD_GasRewardsV2
    foundation_account: CCD_AccountAddress
    pool_parameters: CCD_PoolParametersCpv1
    root_keys: CCD_HigherLevelKeys
    level1_keys: CCD_HigherLevelKeys
    level2_keys: CCD_AuthorizationsV1
    finalization_committee_parameters: CCD_FinalizationCommitteeParameters


class CCD_ChainParameters(BaseModel):
    v0: CCD_ChainParametersV0 = None
    v1: CCD_ChainParametersV1 = None
    v2: CCD_ChainParametersV2 = None


class CCD_InvokeInstanceResponse_Success(BaseModel):
    return_value: bytes
    used_energy: CCD_Energy
    effects: list[CCD_ContractTraceElement]


class CCD_InvokeInstanceResponse_Failure(BaseModel):
    return_value: bytes
    used_energy: CCD_Energy
    reason: CCD_RejectReason


class CCD_InvokeInstanceResponse(BaseModel):
    success: CCD_InvokeInstanceResponse_Success
    failure: CCD_InvokeInstanceResponse_Failure


class CCD_BlockComplete(BaseModel):
    block_info: CCD_BlockInfo
    transaction_summaries: list[CCD_BlockItemSummary]
    special_events: list[CCD_BlockSpecialEvent]
    # This leads to circular import
    # logged_events: Optional[list[MongoTypeLoggedEvent]] = None
    logged_events: Optional[list] = None

    net: str = None



import requests
from rich.console import Console
from ccdexplorer_fundamentals.enums import NET

console = Console()
import sys


class Mixin:
    def ql_get_info_by_accountId(self, value):
        query = "query {"
        query += f'accountByAddress (accountAddress: "{value}")'
        query += """
                    {
                    baker {
                        bakerId
                        state {
                            ... on ActiveBakerState {
                                nodeStatus {
                                    nodeName
                                }
                                pool {
                                    __typename
                                }
                            }
                        }
                    }
                } }
        """
        try:
            # r = self.timed_request(query)
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return r.json()["data"]["accountByAddress"]
            else:
                # self.appriser.notify(title=f"{self.environment}", body=f'query: ql_get_info_by_accountId | value: {value} | error: {r.text}.' )
                console.log(r.text)
                sys.exit(1)
        except Exception as e:
            console.log("Should be sending apprise notification now...")
            # self.appriser.notify(title=f"{self.environment}", body=f'query: ql_get_info_by_accountId | value: {value} | error: {e}.' )
            console.log(query, e)
            sys.exit(1)
            return None

    def ql_get_ql_account_id(self, account_id):
        query = "query {"
        query += f'accountByAddress (accountAddress: "{account_id}")'
        query += """
                    {
                        id
                    }
                }
        """
        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return r.json()["data"]["accountByAddress"]["id"]
            else:
                console.log(r.text)
                sys.exit(1)
        except Exception as e:
            console.log(query, e)
            sys.exit(1)
            return None

    def ql_request_accountInfo_for_account_for_baker(self, account: str):
        query = "query {"
        query += f'accountByAddress (accountAddress: "{account}") '
        query += """
            { 
            address {
                asString
                }
            baker {
                state {
                    ... on ActiveBakerState {
                        pool {
                            openStatus
                            delegatorCount
                            delegatedStake
                            totalStake
                        }
                        pendingChange {
                            ... on PendingBakerReduceStake {
                                effectiveTime
                                newStakedAmount
                            }
                            ... on PendingBakerRemoval {
                                effectiveTime
                            }
                            __typename
                        }
                        stakedAmount
                        restakeEarnings
                    }
                }
            }
            amount
        }
        } # query
        """

        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return r.json()["data"]["accountByAddress"]
            else:
                console.log(r.text)
                sys.exit(1)
        except Exception as e:
            console.log(query, e)
            sys.exit(1)
            return None

    def ql_request_accountInfo_for_account(self, account: str):
        query = "query {"
        query += f'accountByAddress (accountAddress: "{account}") '
        query += """
            { 
            createdAt
            id
            transactionCount
            amount
            address {
            asString
            }
            delegation {
                stakedAmount
                delegatorId
                restakeEarnings
                delegationTarget {
                    ... on BakerDelegationTarget {
                    __typename
                    bakerId
                    }
                    ... on PassiveDelegationTarget {
                    __typename
                    }
                }
            }
            baker {
                state {
                    ... on ActiveBakerState {
                        pool {  
                            delegatorCount
                            delegatedStake
                            totalStake
                            apy_30: apy(period: LAST30_DAYS) {
                            bakerApy
                            delegatorsApy
                            totalApy
                            }
                            apy_7: apy(period: LAST7_DAYS) {
                            bakerApy
                            delegatorsApy
                            totalApy
                            }
                            }
                        stakedAmount
                        restakeEarnings
                    }
                }
            }
            releaseSchedule {
                schedule {
                    nodes {
                        amount
                        timestamp
                        transaction {
                            transactionHash
                        }
                    }
                }
                totalAmount
            }
        }
        } # query
        """

        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return r.json()["data"]["accountByAddress"]
            else:
                sys.exit(1)
                return None

        except Exception as e:
            sys.exit(1)
            console.log(query, e)
            return None

    def ql_request_transactions_for_lookup(
        self,
        accountLookup: bool,
        lookup_value: str,
        before: str = None,
        after: str = None,
        net=NET.MAINNET,
    ):
        query = "query {"
        if accountLookup:
            query += f'accountByAddress (accountAddress: "{lookup_value}") '
        else:  # blocklookup
            query += f'blockByBlockHash (blockHash: "{lookup_value}") '

        query += "{ transactionCount "
        if before == "first":
            accounts_str = f"transactions (first: {self.nodes_request_limit}) "

        elif before != "null":
            accounts_str = (
                f'transactions (last: {self.nodes_request_limit}, before: "{before}") '
            )
        elif "last" in after:
            accounts_str = f"transactions ({after}) "

        elif after != "null":
            accounts_str = (
                f'transactions (first: {self.nodes_request_limit}, after: "{after}") '
            )
        else:
            accounts_str = f"transactions (first: {self.nodes_request_limit})"
        query += accounts_str + "{"
        query += self.pageInfo()
        if accountLookup:
            query += "nodes { transaction {"
        else:
            query += "nodes { "

        query += self.standard_tx_fields()
        query += self.ql_query_tx_events()

        if accountLookup:
            query += "} } }  } "
        else:
            query += "} }  } "

        try:
            url_to_use = (
                self.graphql_url if net == NET.MAINNET else self.graphql_url_testnet
            )
            r = requests.post(url_to_use, json={"query": query})
            if r.status_code == 200:
                if accountLookup:
                    return (
                        r.json()["data"]["accountByAddress"]["transactions"]["nodes"],
                        r.json()["data"]["accountByAddress"]["transactions"][
                            "pageInfo"
                        ],
                        r.json()["data"]["accountByAddress"]["transactionCount"],
                    )
                else:
                    return (
                        r.json()["data"]["blockByBlockHash"]["transactions"]["nodes"],
                        r.json()["data"]["blockByBlockHash"]["transactions"][
                            "pageInfo"
                        ],
                        r.json()["data"]["blockByBlockHash"]["transactionCount"],
                    )
            else:
                console.log(r.text)
                console.log(query)
                sys.exit(1)

        except Exception as e:
            console.log(r.text)
            console.log(query)
            sys.exit(1)
            console.log(query, e)
            return [], {}, 0

    def ql_request_account_rewards_for_lookup(
        self, lookup_value: str, before: str = None, after: str = None
    ):
        LIMIT = self.nodes_request_limit
        LIMIT = 21
        query = "query {"
        query += f'  accountByAddress (accountAddress: "{lookup_value}") {{ '

        if before == "first":
            accounts_str = f"rewards (first: {LIMIT}) "

        elif before != "null":
            accounts_str = f'rewards (last: {LIMIT}, before: "{before}") '
        elif "last" in after:
            accounts_str = f"rewards ({after}) "

        elif after != "null":
            accounts_str = f'rewards (first: {LIMIT}, after: "{after}") '
        else:
            accounts_str = f"rewards (first: {LIMIT})"
        query += accounts_str + "{"
        query += self.pageInfo()
        query += """
            nodes {
                amount
                timestamp
                __typename
                rewardType
            }
            } } } 
        """

        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return (
                    r.json()["data"]["accountByAddress"]["rewards"]["nodes"],
                    r.json()["data"]["accountByAddress"]["rewards"]["pageInfo"],
                )

            else:
                console.log(r.text)
                sys.exit(1)
        except Exception as e:
            console.log(query, e)
            sys.exit(1)
            return None

    def ql_request_transactions_for_lookup_using_edges(
        self, lookup_value: str, before_cursor: str = None, after: str = None
    ):
        query = "query {"
        query += f'accountByAddress (accountAddress: "{lookup_value}") '

        query += "{ "
        if not before_cursor:
            if not after:
                accounts_str = f"transactions (first: 50) "
            else:
                accounts_str = f'transactions (first: 50, after: "{after}") '
        else:
            accounts_str = f'transactions (last: 50, before: "{before_cursor}") '

        query += accounts_str + "{"
        query += self.pageInfo()
        query += "edges { before_cursor: cursor  node { transaction {"

        query += self.standard_tx_fields()
        query += self.ql_query_tx_events()

        query += "} } }  } } "

        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return (
                    r.json()["data"]["accountByAddress"]["transactions"]["edges"],
                    r.json()["data"]["accountByAddress"]["transactions"]["pageInfo"],
                )
            else:

                console.log(r.text, query)
                return [], {"hasNextPage": False}

        except Exception as e:
            console.log(query, e)
            return [], {"hasNextPage": False}

    def ql_request_delegators_for_lookup_using_edges(
        self, baker_id: str, before_cursor: str = None, after: str = None
    ):
        query = "query {"
        query += f"bakerByBakerId (bakerId: {baker_id}) "
        query += "{ state { ... on ActiveBakerState { pool { delegatorCount "
        # query += '{ '
        if not before_cursor:
            if not after:
                accounts_str = f"delegators (first: 50) "
            else:
                accounts_str = f'delegators (first: 50, after: "{after}") '
        else:
            accounts_str = f'delegators (last: 50, before: "{before_cursor}") '

        query += accounts_str + "{"
        query += self.pageInfo()
        query += "edges { before_cursor: cursor  node { "
        query += """
                            stakedAmount
                            restakeEarnings
                            accountAddress {
                                asString
                            }
                        }
            """
        # query += self.standard_tx_fields()
        # query += self.ql_query_tx_events()

        query += "} } }  } } } }"

        try:
            r = requests.post(self.graphql_url, json={"query": query})
            if r.status_code == 200:
                return (
                    r.json()["data"]["bakerByBakerId"]["state"]["pool"]["delegators"][
                        "edges"
                    ],
                    r.json()["data"]["bakerByBakerId"]["state"]["pool"]["delegators"][
                        "pageInfo"
                    ],
                )
            else:
                console.log(r.text, query)
                return [], {"pageInfo": False}

        except Exception as e:
            console.log(query, e)
            return [], {"pageInfo": False}



from enum import Enum
from ccdexplorer_fundamentals.env import (
    ADMIN_CHAT_ID,
    MAILTO_LINK,
    MAILTO_USER,
    NOTIFIER_API_TOKEN,
    API_TOKEN,
    FASTMAIL_TOKEN,
    BRANCH,
    ENVIRONMENT,
)
import requests


class TooterType(Enum):
    INFO = "info"
    GRAPHQL_ERROR = "graphql error"
    MONGODB_ERROR = "mongodb error"
    REQUESTS_ERROR = "requests error"
    BOT_MAIN_LOOP_ERROR = "bot_main_loop_error"


class TooterChannel(Enum):
    BOT = "bot"
    NOTIFIER = "notifier"
    SMART = "smart"


class Tooter:
    def __init__(self) -> None:
        self.environment = ENVIRONMENT
        self.branch = BRANCH
        self.NOTIFIER_API_TOKEN = NOTIFIER_API_TOKEN
        self.BOT_API_TOKEN = API_TOKEN
        self.FASTMAIL_TOKEN = FASTMAIL_TOKEN
        self.url = "https://tooter.ccdexplorer.io/notify/"
        self.email_part_1 = MAILTO_LINK
        self.email_part_2 = f"""&user={MAILTO_USER}&pass={FASTMAIL_TOKEN}&from=CCDExplorer Bot<bot@ccdexplorer.io>"""

    def email(self, title: str, body: str, email_address: str, value=None, error=None):
        body_signature = """
Please visit your <a href='https://ccdexplorer.io/settings/user/overview'>account</a> to adjust notification settings.

        """
        payload = {
            "urls": f"{self.email_part_1}{email_address}{self.email_part_2 }",
            "title": title,
            "body": body + body_signature,
            "format": "html",
        }
        _ = requests.post(self.url, json=payload)

    def relay(
        self,
        channel: TooterChannel,
        body: str,
        notifier_type: TooterType,
        title: str,
        chat_id: int = None,
        value=None,
        error=None,
        bcc=False,
    ):
        API_TO_USE = (
            self.BOT_API_TOKEN
            if channel == TooterChannel.BOT
            else self.NOTIFIER_API_TOKEN
        )
        chat_id = ADMIN_CHAT_ID if channel == TooterChannel.NOTIFIER else chat_id

        payload = {
            "urls": f"tgram://{API_TO_USE}/{chat_id}",
            "title": f"{title}<br/>",
            "body": body,
            "format": "html",
        }
        _ = requests.post(self.url, json=payload)

    def send(
        self,
        channel: TooterChannel,
        message: str,
        notifier_type: TooterType,
        chat_id: int = None,
        value=None,
        error=None,
    ):
        API_TO_USE = (
            self.BOT_API_TOKEN
            if channel == TooterChannel.BOT
            else self.NOTIFIER_API_TOKEN
        )
        chat_id = ADMIN_CHAT_ID if channel == TooterChannel.NOTIFIER else chat_id

        title = f"t: {notifier_type.value} | e: {self.environment} | b: {self.branch}"
        body = message
        if value:
            body += "| value: {value} "
        if error:
            body += "| error: {error}."

        payload = {
            "urls": f"tgram://{API_TO_USE}/{chat_id}",
            "title": title,
            "body": body,
            "format": "html",
        }
        _ = requests.post(self.url, json=payload)
